@@ -1,6 +1,6 @@
 /*
  *	Copyright © Elvis Pfützenreuter, 2000
- *	Copyright © Jan Engelhardt, 2006 - 2008
+ *	Copyright © Jan Engelhardt, 2006 - 2009
  *	Copyright © Bastian Kleineidam, 2005
  *
  *	This file is part of pam_mount; you can redistribute it and/or
@@ -13,7 +13,6 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <assert.h>
-#include <ctype.h>
 #include <errno.h>
 #include <limits.h>
 #include <signal.h>
@@ -22,35 +21,16 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <libHX/ctype_helper.h>
 #include <libHX/defs.h>
-#include <libHX.h>
+#include <libHX/deque.h>
+#include <libHX/proc.h>
 #include <grp.h>
 #include <pwd.h>
-#include "misc.h"
-#include "mount.h"
 #include "pam_mount.h"
-#include "readconfig.h"
-#include "spawn.h"
-#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
-#	include <fstab.h>
-#elif defined(__linux__)
-#	include <mntent.h>
-#	include <sys/ioctl.h>
-#	include <fcntl.h>
-#	include <linux/loop.h>
-#	include <linux/major.h>
-#endif
-#ifndef S_IXUGO
-#	define S_IXUGO (S_IXUSR | S_IXGRP | S_IXOTH)
-#endif
 
 /* Functions */
-static int already_mounted(const struct config * const, const struct vol *, struct HXbtree *);
-static int fstype_nodev(const char *);
 static inline bool mkmountpoint(struct vol *, const char *);
-static int pipewrite(int, const void *, size_t);
-static void run_ofl(const struct config * const, struct HXbtree *);
-static hxmc_t *vol_to_dev(const struct vol *);
 
 //-----------------------------------------------------------------------------
 /**
@@ -66,24 +46,29 @@ static hxmc_t *vol_to_dev(const struct vol *);
  */
 static void log_output(int fd, const char *cmsg)
 {
-	char buf[BUFSIZ + 1];
+	hxmc_t *buf = NULL;
 	FILE *fp;
 
 	if ((fp = fdopen(fd, "r")) == NULL) {
 		w4rn("error opening file: %s\n", strerror(errno));
+		close(fd);
 		return;
 	}
 
 	setvbuf(fp, NULL, _IOLBF, 0);
-	if (fgets(buf, sizeof(buf), fp) != NULL) {
-		if (cmsg != NULL)
-			w4rn("%s", cmsg);
+	do {
+		if (HX_getl(&buf, fp) == NULL)
+			break;
+		HX_chomp(buf);
+		if (*buf != '\0' && cmsg != NULL) {
+			l0g("%s", cmsg);
+			cmsg = NULL;
+		}
 
-		do {
-			w4rn("%s", buf);
-		} while (fgets(buf, sizeof(buf), fp) != NULL);
-	}
+		l0g("%s\n", buf);
+	} while (true);
 	fclose(fp);
+	HXmc_free(buf);
 }
 
 /**
@@ -92,18 +77,37 @@ static void log_output(int fd, const char *cmsg)
  * @vinfo:
  *
  * Runs `ofl` on a directory/mountpoint and logs its output, for debugging
- * purposes.
+ * purposes. (ofl is a better-suited lsof/fuser.)
  */
-static void run_ofl(const struct config *const config, struct HXbtree *vinfo)
+static void run_ofl(const struct config *const config, const char *mntpt,
+    unsigned int signum)
 {
-	const char *mntpt;
+	struct HXformat_map *vinfo;
+	struct HXproc proc;
+	struct HXdeque *argv;
 	struct stat sb;
+	int ret;
 
-	mntpt = HXbtree_get(vinfo, "MNTPT");
 	if (stat(mntpt, &sb) < 0 && errno == ENOENT)
 		return;
-	ofl_printf = misc_warn;
-	ofl(mntpt, 0);
+
+	vinfo = HXformat_init();
+	if (vinfo == NULL)
+		return;
+	format_add(vinfo, "MNTPT", mntpt);
+	HXformat_add(vinfo, "SIGNAL", reinterpret_cast(void *,
+		static_cast(long, signum)), HXFORMAT_IMMED | HXTYPE_UINT);
+	argv = arglist_build(config->command[CMD_OFL], vinfo);
+	HXformat_free(vinfo);
+	if (argv == NULL)
+		return;
+	memset(&proc, 0, sizeof(proc));
+	proc.p_flags = HXPROC_VERBOSE;
+	ret = pmt_spawn_dq(argv, &proc);
+	if (ret <= 0)
+		l0g("error executing ofl: %s\n", strerror(-ret));
+	else
+		HXproc_wait(&proc);
 }
 
 /**
@@ -115,139 +119,39 @@ static void run_ofl(const struct config *const config, struct HXbtree *vinfo)
  * Checks if @config->volume[@vol] is already mounted, and returns 1 if this
  * the case, 0 if not and -1 on error.
  */
-static int already_mounted(const struct config *const config,
-    const struct vol *vpt, struct HXbtree *vinfo)
-#if defined(__linux__)
-{
-	hxmc_t *dev;
-	char real_mpt[PATH_MAX+1];
-	struct mntent *mtab_record;
-	bool mounted = false;
-	FILE *mtab;
-
-	if ((dev = vol_to_dev(vpt)) == NULL) {
-		l0g("pmt::vol_to_dev: %s\n", strerror(errno));
-		return -1;
-	}
-
-	if ((mtab = setmntent("/etc/mtab", "r")) == NULL) {
-		l0g("could not open /etc/mtab\n");
-		HXmc_free(dev);
-		return -1;
-	}
-	if (realpath(vpt->mountpoint, real_mpt) == NULL) {
-		w4rn("can't get realpath of volume %s: %s\n",
-		     vpt->mountpoint, strerror(errno));
-		strncpy(real_mpt, vpt->mountpoint, sizeof_z(real_mpt));
-		real_mpt[sizeof_z(real_mpt)] = '\0';
-	} else {
-		real_mpt[sizeof_z(real_mpt)] = '\0';
-		l0g("realpath of volume \"%s\" is \"%s\"\n",
-		    vpt->mountpoint, real_mpt);
-	}
-
-	w4rn("checking to see if %s is already mounted at %s\n",
-	     dev, vpt->mountpoint);
-
-	while ((mtab_record = getmntent(mtab)) != NULL) {
-		const char *fsname = mtab_record->mnt_fsname;
-		const char *fstype = mtab_record->mnt_type;
-		const char *fspt   = mtab_record->mnt_dir;
-		int (*xcmp)(const char *, const char *);
-		struct loop_info64 loopdev;
-		struct stat statbuf;
-
-#ifdef HAVE_STRUCT_LOOP_INFO64_LO_FILE_NAME
-		if (stat(fsname, &statbuf) == 0 && S_ISBLK(statbuf.st_mode) &&
-		    major(statbuf.st_rdev) == LOOP_MAJOR)
-			/*
-			 * If /etc/mtab is a link to /proc/mounts then the loop
-			 * device instead of the real device will be listed --
-			 * resolve it.
-			 */
-			fsname = loop_file_name(fsname, &loopdev);
-#endif
-
-		xcmp = (strcmp(fstype, "smbfs") == 0 ||
-		        strcmp(fstype, "cifs") == 0 ||
-		        strcmp(fstype, "ncpfs") == 0) ? strcasecmp : strcmp;
-
-		if (xcmp(fsname, dev) == 0 &&
-		    (strcmp(fspt, vpt->mountpoint) == 0 ||
-		    strcmp(fspt, real_mpt) == 0)) {
-			mounted = true;
-			break;
-		}
-	}
-
-	endmntent(mtab);
-	HXmc_free(dev);
-	return mounted;
-}
-#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
-{
-	hxmc_t *dev;
-	struct HXdeque *argv;
-	char mte[BUFSIZ + 1];
-	int i, cstdout = -1, mounted = 0;
-	struct vol *vpt;
-	pid_t pid;
-	FILE *fp;
-
-	vpt = &config->volume[vol];
-	if ((dev = vol_to_dev(vpt)) == NULL) {
-		l0g("pmt::vol_to_dev: %s\n", strerror(errno));
-		return -1;
-	}
-
-	//getmntinfo() or getfsstat()
-	while (...) {
-		struct statfs *sb;
-		/* FIXME: Test it. */
-		int (*xcmp)(const char *, const char *);
-
-		/* 
-		 * Use case-insensitive for SMB, etc. FIXME: Is it called
-		 * "smbfs" under BSD too?
-		 */
-		xcmp = (sb->f_fstypename != NULL &&
-		       (strcmp(sb->f_fstypename, "smbfs") == 0 ||
-		       strcmp(sb->f_fstypename, "cifs") == 0 ||
-		       strcmp(sb->f_fstypename, "ncpfs") == 0)) ?
-		       strcasecmp : strcmp;
-
-		/*
-		 * FIXME: Does BSD also turn "symlink mountpoints" into "real
-		 * mountpoints"?
-		 */
-		if (xcmp(sb->f_mntfromname, dev) == 0 &&
-		    strcmp(sb->f_mntonname, vpt->mountpoint) == 0) {
-			mounted = 1;
-			break;
-		}
-	}
-
-	fclose(fp); /* automatically closes @cstdout */
-	if (waitpid(pid, NULL, 0) != 0)
-		l0g("error waiting for child: %s\n", strerror(errno));
-	spawn_restore_sigchld();
-	HXmc_free(dev);
-	return mounted;
-}
+#if defined(HAVE_GETMNTENT)
+	/* elsewhere */
+#elif defined(HAVE_GETMNTINFO)
+	/* elsewhere */
 #else
+int pmt_already_mounted(const struct config *const config,
+    const struct vol *vpt, struct HXformat_map *vinfo)
 {
 	l0g("check for previous mount not implemented on arch.\n");
 	return -1;
 }
 #endif
 
+static bool fstype_networked(enum command_type fstype)
+{
+	switch (fstype) {
+	case CMD_NFSMOUNT:
+	case CMD_CIFSMOUNT:
+	case CMD_SMBMOUNT:
+	case CMD_NCPMOUNT:
+		return true;
+	default:
+		return false;
+	}
+}
+
 /**
  * vol_to_dev -
  * @vol:	volume to analyze
  *
  * Turn a volume into the mountspec as accepted by the specific mount program.
  */
-static hxmc_t *vol_to_dev(const struct vol *vol)
+hxmc_t *pmt_vol_to_dev(const struct vol *vol)
 {
 	hxmc_t *ret;
 
@@ -273,6 +177,14 @@ static hxmc_t *vol_to_dev(const struct vol *vol)
 		break;
 
 	default:
+		if (!fstype_networked(vol->type) && vol->server != NULL)
+			/*
+			 * Possible causes: we do not know about the fs yet.
+			 * (Was once the case with NFS4, for example.)
+			 */
+			l0g("The \"server\" attribute is ignored for this "
+			    "filesystem (%s).\n", vol->fstype);
+
 		ret = HXmc_strinit(vol->volume);
 		break;
 	}
@@ -287,71 +199,23 @@ static void log_pm_input(const struct config *const config,
 
 	options = kvplist_to_str(&vpt->options);
 	w4rn(
-		"Mount info: %s, user=%s <volume server=\"%s\" path=\"%s\" "
+		"Mount info: %s, user=%s <volume fstype=\"%s\" "
+		"server=\"%s\" path=\"%s\" "
 		"mountpoint=\"%s\" cipher=\"%s\" fskeypath=\"%s\" "
 		"fskeycipher=\"%s\" fskeyhash=\"%s\" options=\"%s\" /> "
-		"fstab=%u\n",
+		"fstab=%u ssh=%u\n",
 		vpt->globalconf ? "globalconf" : "luserconf",
-		znul(vpt->user), znul(vpt->server), znul(vpt->volume),
+		znul(vpt->user), znul(vpt->fstype),
+		znul(vpt->server), znul(vpt->volume),
 		vpt->mountpoint, znul(vpt->cipher), znul(vpt->fs_key_path),
 		znul(vpt->fs_key_cipher), znul(vpt->fs_key_hash), options,
-		vpt->use_fstab
+		vpt->use_fstab, vpt->uses_ssh
 	);
 	HXmc_free(options);
 }
 
 /**
- * mkmountpoint_real - create mountpoint directory
- * @volume:	volume description
- * @d:		directory
- *
- * If the directory @d does not exist, create it and all its parents if
- * @volume->created_mntpt = true. On success, returns true, otherwise false.
- */
-static bool mkmountpoint_real(struct vol *const volume, const char *const d)
-{
-	bool ret = true;
-	struct passwd *passwd_ent;
-	char dcopy[PATH_MAX + 1], *parent;
-
-	assert(d != NULL);
-
-	strncpy(dcopy, d, sizeof_z(dcopy));
-	dcopy[sizeof_z(dcopy)] = '\0';
-	parent = HX_dirname(dcopy);
-	if (!pmt_fileop_exists(parent) && mkmountpoint(volume, parent) == 0) {
-		ret = false;
-		goto out;
-	}
-	if ((passwd_ent = getpwnam(volume->user)) == NULL) {
-		l0g("could not determine uid from %s to make %s\n", volume->user, d);
-		ret = false;
-		goto out;
-	}
-	/*
-	 * The directory will be created in a restricted mode S_IRWXU here.
-	 * When mounted, the root directory of the new vfsmount will override
-	 * it, so there is no need to use S_IRWXUGO or S_IRWXU | S_IXUGO here.
-	 *
-	 * Workaround for CIFS on root_squashed NFS: +S_IXUGO
-	 */
-	if (mkdir(d, S_IRWXU | S_IXUGO) < 0) {
-		ret = false;
-		goto out;
-	}
-	if (chown(d, passwd_ent->pw_uid, passwd_ent->pw_gid) < 0) {
-		l0g("could not chown %s to %s\n", d, volume->user);
-		ret = false;
-		goto out;
-	}
-	volume->created_mntpt = true;
- out:
-	free(parent);
-	return ret;
-}
-
-/**
- * mkmountpoint_pick - create mountpoint for volume
+ * mkmountpoint - create mountpoint for volume
  * @volume:	volume structure
  * @d:		directory to create
  *
@@ -362,40 +226,79 @@ static bool mkmountpoint_real(struct vol *const volume, const char *const d)
  *
  * If that fails, do as usual (create as root, chown to user).
  */
-static bool mkmountpoint_pick(struct vol *volume, const char *d)
+static bool mkmountpoint(struct vol *volume, const char *d)
 {
-	struct passwd *pe;
-	bool ret;
+	const struct passwd *pe;
+	hxmc_t *dtmp;
+	char *last;
+	bool ret = true;
 
 	if ((pe = getpwnam(volume->user)) == NULL) {
 		l0g("getpwuid: %s\n", strerror(errno));
 		return false;
 	}
+	dtmp = HXmc_strinit(d);
+	if (dtmp == NULL || HXmc_strcat(&dtmp, "/") == NULL) {
+		l0g("HXmc_strinit: %s\n", strerror(errno));
+		return false;
+	}
 
-	w4rn("creating mount point %s\n", d);
-	if (seteuid(pe->pw_uid) == 0)
-		if (mkmountpoint_real(volume, d))
-			return true;
-
+	last = dtmp;
+	while ((last = strchr(last + 1, '/')) != NULL) {
+		*last = '\0';
+		w4rn("%s: checking %s\n", __func__, dtmp);
+		if (pmt_fileop_exists(dtmp)) {
+			/* If it exists, all is good... */
+			*last = '/';
+			continue;
+		}
+		/*
+		 * Prefer creation using user id, due to NFS possibly using
+		 * root_squashed NFS.
+		 *
+		 * The directory will be created in a restricted mode S_IRWXU
+		 * here. When mounted, the root directory of the new vfsmount
+		 * will override it, so there is no need to use S_IRWXUGO or
+		 * S_IRWXU | S_IXUGO here.
+		 *
+		 * Workaround for CIFS on root_squashed NFS: +S_IXUGO
+		 */
+		if (setegid(pe->pw_gid) < 0 ||
+		    seteuid(pe->pw_uid) < 0) {
+			l0g("seteuid/setegid %ld:%ld failed: %s\n",
+			    static_cast(long, pe->pw_uid),
+			    static_cast(long, pe->pw_gid), strerror(errno));
+		} else if (mkdir(dtmp, S_IRWXU | S_IXUGO) == 0) {
+			/* Was createable as user - ok */
+			w4rn("mkdir[%ld] %s\n",
+			     static_cast(long, pe->pw_uid), dtmp);
+			*last = '/';
+			continue;
+		}
+		/* Try with root again. */
+		if (seteuid(0) < 0) {
+			l0g("seteuid 0 failed\n");
+			ret = false;
+			break;
+		}
+		if (mkdir(dtmp, S_IRWXU | S_IXUGO) < 0) {
+			l0g("mkdir %s failed: %s\n", dtmp, strerror(errno));
+			ret = false;
+			break;
+		}
+		w4rn("mkdir[0] %s\n", dtmp);
+		if (chown(dtmp, pe->pw_uid, pe->pw_gid) < 0) {
+			l0g("chown %s failed: %s\n", dtmp, strerror(errno));
+			ret = false;
+			break;
+		}
+		w4rn("chown %s -> %ld:%ld\n", dtmp,
+		     pe->pw_uid, pe->pw_gid);
+	}
+	HXmc_free(dtmp);
+	/* restore state: */
 	seteuid(0);
-	ret = mkmountpoint_real(volume, d);
-	if (!ret)
-		l0g("tried to create %s but failed\n", d);
-	return ret;
-}
-
-/**
- * mkmountpoint -
- *
- * Wrapper for mkmountpoint_pick(). Switch back to root user after
- * mkmountpoint() operation. This is needed, otherwise the PAM stack will
- * (more or less) spuriously fail with PAM_SYSTEM_ERR.
- */
-static inline bool mkmountpoint(struct vol *volume, const char *d)
-{
-	bool r = mkmountpoint_pick(volume, d);
-	seteuid(0);
-	return r;
+	return volume->created_mntpt = ret;
 }
 
 /**
@@ -408,12 +311,11 @@ static inline bool mkmountpoint(struct vol *volume, const char *d)
  * Returns zero on error, positive non-zero for success.
  */
 int do_unmount(const struct config *config, struct vol *vpt,
-    struct HXbtree *vinfo, const char *const password)
+    struct HXformat_map *vinfo, const char *const password)
 {
 	struct HXdeque *argv;
-	int child_exit, ret = 1, cstderr = -1;
-	pid_t pid = -1;
-	int type;
+	struct HXproc proc;
+	int ret, type;
 
 	assert(vinfo != NULL);
 	assert(password == NULL);	/* password should point to NULL for unmounting */
@@ -421,9 +323,9 @@ int do_unmount(const struct config *config, struct vol *vpt,
 	if (Debug)
 		/*
 		 * Often, a process still exists with ~ as its pwd after
-		 * logging out.  Running lsof helps debug this.
+		 * logging out. Running ofl helps debug this.
 		 */
-		run_ofl(config, vinfo);
+		run_ofl(config, vpt->mountpoint, 0);
 
 	switch (vpt->type) {
 		case CMD_CRYPTMOUNT:
@@ -447,23 +349,20 @@ int do_unmount(const struct config *config, struct vol *vpt,
 		l0g("{smb,ncp}umount not defined in pam_count.conf.xml\n");
 
 	argv = arglist_build(config->command[type], vinfo);
-	if (!spawn_start(argv, &pid, NULL, NULL, &cstderr, set_myuid, NULL)) {
+	memset(&proc, 0, sizeof(proc));
+	proc.p_flags = HXPROC_VERBOSE | HXPROC_NULL_STDOUT | HXPROC_STDERR;
+	proc.p_ops   = &pmt_dropprivs_ops;
+	if ((ret = pmt_spawn_dq(argv, &proc)) <= 0) {
 		ret = 0;
 		goto out;
 	}
 
-	log_output(cstderr, "umount errors:\n");
-	w4rn("waiting for umount\n");
-	if (waitpid(pid, &child_exit, 0) < 0) {
-		l0g("error waiting for child: %s\n", strerror(errno));
-		ret = 0;
-		goto out;
-	} else {
+	log_output(proc.p_stderr, "umount messages:\n");
+	if ((ret = HXproc_wait(&proc)) >= 0)
 		/* pass on through the result from the umount process */
-		ret = !WEXITSTATUS(child_exit);
-	}
+		ret = proc.p_exited && proc.p_status == 0;
+
  out:
-	spawn_restore_sigchld();
 	if (config->mkmntpoint && config->rmdir_mntpt && vpt->created_mntpt)
 		if (rmdir(vpt->mountpoint) < 0)
 			/* non-fatal, but warn */
@@ -471,38 +370,8 @@ int do_unmount(const struct config *config, struct vol *vpt,
 	return ret;
 }
 
-/* INPUT: fd, a valid file descriptor; buf, a buffer of size count
- * SIDE EFFECTS: buf is written to fd
- * OUTPUT: number of bytes written or 0 on error
- * NOTE: SIGPIPE is ignored during this operation to avoid "broken pipe"
- */
-static int pipewrite(int fd, const void *buf, size_t count)
-{
-	struct sigaction ignoresact = {.sa_handler = SIG_IGN}, oldsact;
-	int fnval;
-
-	assert(fd >= 0);
-	assert(buf != NULL);
-	assert(count >= 0);
-
-	/* avoid bomb on command exiting before data written */
-	sigemptyset(&ignoresact.sa_mask);
-	if (sigaction(SIGPIPE, &ignoresact, &oldsact) < 0) {
-		fnval = -1;
-		goto out;
-	}
-	fnval = write(fd, buf, count);
-	/* restore old handler */
-	if (sigaction(SIGPIPE, &oldsact, NULL) < 0) {
-		fnval = -1;
-		goto out;
-	}
- out:
-	return fnval;
-}
-
 static int check_filesystem(const struct config *config, const struct vol *vpt,
-    struct HXbtree *vinfo)
+    struct HXformat_map *vinfo)
 {
 /* PRE:    config points to a valid struct config
  *         config->volume[vol] is a valid struct vol
@@ -510,10 +379,10 @@ static int check_filesystem(const struct config *config, const struct vol *vpt,
  * FN VAL: if error 0 else 1, errors are logged
  */
 #if defined (__linux__)
-	pid_t pid;
-	int child_exit, cstdout = -1, cstderr = -1;
 	const char *fsck_target;
 	struct HXdeque *argv;
+	struct HXproc proc;
+	int ret;
 
 	assert(vinfo != NULL);
 
@@ -541,21 +410,23 @@ static int check_filesystem(const struct config *config, const struct vol *vpt,
 	format_add(vinfo, "FSCKTARGET", fsck_target);
 
 	argv = arglist_build(config->command[CMD_FSCK], vinfo);
-	if (!spawn_start(argv, &pid, NULL, &cstdout, &cstderr, NULL, NULL))
+	memset(&proc, 0, sizeof(proc));
+	proc.p_flags = HXPROC_VERBOSE | HXPROC_STDOUT | HXPROC_STDERR;
+	if ((ret = pmt_spawn_dq(argv, &proc)) <= 0)
 		return 0;
 
 	/* stdout and stderr must be logged for fsck */
-	log_output(cstdout, NULL);
-	log_output(cstderr, NULL);
+	log_output(proc.p_stdout, NULL);
+	log_output(proc.p_stderr, NULL);
 	w4rn("waiting for filesystem check\n");
-	if (waitpid(pid, &child_exit, 0) < 0)
-		l0g("error waiting for child: %s\n", strerror(errno));
-	spawn_restore_sigchld();
+	if ((ret = HXproc_wait(&proc)) < 0)
+		l0g("error waiting for child: %s\n", strerror(-ret));
+
 	/*
 	 * pass on through the result -- okay if 0 (no errors) or 1 (errors
 	 * corrected)
 	 */
-	return WEXITSTATUS(child_exit) == 0 || WEXITSTATUS(child_exit) == 1;
+	return proc.p_exited && (proc.p_status == 0 || proc.p_status == 1);
 #else
 	l0g("checking filesystem not implemented on arch.\n");
 	return 1;
@@ -569,7 +440,7 @@ static int check_filesystem(const struct config *config, const struct vol *vpt,
  * @vinfo:	variable substituions
  */
 static void mount_set_fsck(const struct config *config,
-    const struct vol *vol, struct HXbtree *vinfo)
+    const struct vol *vol, struct HXformat_map *vinfo)
 {
 	const struct HXdeque_node *i;
 	hxmc_t *string, *current;
@@ -581,7 +452,7 @@ static void mount_set_fsck(const struct config *config,
 	string = HXmc_meminit(NULL, 0);
 
 	for (i = config->command[CMD_FSCK]->first; i != NULL; i = i->next) {
-		if (HXformat_aprintf(vinfo, &current, i->ptr) > 0) {
+		if (HXformat2_aprintf(vinfo, &current, i->ptr) > 0) {
 			HXmc_strcat(&string, current);
 			HXmc_strcat(&string, " ");
 		}
@@ -597,29 +468,27 @@ static void mount_set_fsck(const struct config *config,
  * @config:	current config
  * @vpt:	volume descriptor
  * @vinfo:
- * @password:	login password
+ * @password:	login password (may be %NULL)
  *
  * Returns zero on error, positive non-zero for success.
  */
 int do_mount(const struct config *config, struct vol *vpt,
-    struct HXbtree *vinfo, const char *password)
+    struct HXformat_map *vinfo, const char *password)
 {
 	const struct HXdeque_node *n;
 	struct HXdeque *argv;
-	int child_exit = 0, cstdin = -1, cstderr = -1;
+	struct HXproc proc;
 	const char *mount_user;
-	pid_t pid = -1;
 	int ret;
 
 	assert(vinfo != NULL);
-	assert(password != NULL);
 
-	ret = already_mounted(config, vpt, vinfo);
-	if (ret == -1) {
+	ret = pmt_already_mounted(config, vpt, vinfo);
+	if (ret < 0) {
 		l0g("could not determine if %s is already mounted, "
 		    "failing\n", vpt->volume);
 		return 0;
-	} else if (ret == 1) {
+	} else if (ret > 0) {
 		w4rn("%s already seems to be mounted at %s, "
 		     "skipping\n", vpt->volume, vpt->mountpoint);
 		return 1;
@@ -641,10 +510,8 @@ int do_mount(const struct config *config, struct vol *vpt,
 		    "pam_mount.conf.xml\n");
 		return 0;
 	}
-	w4rn("checking for encrypted filesystem key configuration\n");
 
 	password = (password != NULL) ? password : "";
-	w4rn("about to start building mount command\n");
 	if ((argv = HXdeque_init()) == NULL)
 		misc_log("malloc: %s\n", strerror(errno));
 	if (vpt->uses_ssh)
@@ -655,6 +522,12 @@ int do_mount(const struct config *config, struct vol *vpt,
 	for (n = config->command[vpt->type]->first; n != NULL; n = n->next)
 		arglist_add(argv, n->ptr, vinfo);
 
+	/*
+	 * Note to future editors: Do not do a second-time substitution of the
+	 * arguments. Variables specified within <volume>s are already expanded
+	 * in expandconfig(), but see the comment in mount_op().
+	 */
+
 	if (vpt->type == CMD_LCLMOUNT &&
 	    !check_filesystem(config, vpt, vinfo))
 		l0g("error checking filesystem but will continue\n");
@@ -665,31 +538,44 @@ int do_mount(const struct config *config, struct vol *vpt,
 	mount_set_fsck(config, vpt, vinfo);
 	arglist_log(argv);
 	mount_user = vpt->noroot ? vpt->user : NULL;
-	if (!spawn_start(argv, &pid, &cstdin, NULL, &cstderr,
-	    set_myuid, mount_user))
+	memset(&proc, 0, sizeof(proc));
+	proc.p_flags = HXPROC_VERBOSE | HXPROC_STDIN |
+	               HXPROC_NULL_STDOUT | HXPROC_STDERR;
+	proc.p_ops   = &pmt_dropprivs_ops;
+	proc.p_data  = const_cast1(char *, mount_user);
+	if ((ret = pmt_spawn_dq(argv, &proc)) <= 0)
 		return 0;
 
-	if (vpt->type != CMD_NFSMOUNT)
-		if (pipewrite(cstdin, password, strlen(password)) !=
-		    strlen(password))
-			/* FIXME: clean: returns value of exit below */
-			l0g("error sending password to mount\n");
-	close(cstdin);
+	if (write(proc.p_stdin, password, strlen(password)) !=
+	    strlen(password))
+		/* FIXME: clean: returns value of exit below */
+		l0g("error sending password to mount\n");
+	close(proc.p_stdin);
 
-	log_output(cstderr, "mount errors:\n");
-	w4rn("waiting for mount\n");
-	if (waitpid(pid, &child_exit, 0) < 0) {
-		spawn_restore_sigchld();
-		l0g("error waiting for child: %s\n", strerror(errno));
+	log_output(proc.p_stderr, "Messages from underlying mount program:\n");
+	if ((ret = HXproc_wait(&proc)) < 0) {
+		l0g("error waiting for child: %s\n", strerror(-ret));
 		return 0;
 	}
 
-	spawn_restore_sigchld();
+	if (!proc.p_exited || proc.p_status != 0) {
+		/*
+		 * Remove mountpoint if mount failed, to flag unavailability
+		 * of service (e.g. when mntpt is the user's home directory).
+		 */
+		if (rmdir(vpt->mountpoint) < 0)
+			/* non-fatal, but warn */
+			w4rn("could not remove %s again: %s\n",
+			     vpt->mountpoint, strerror(errno));
+		vpt->created_mntpt = false;
+	}
+
 	if (Debug)
-		spawn_synchronous((const char *const []){"df", "-Ta", NULL});
+		HXproc_run_sync((const char *const []){"df", "-Ta", NULL},
+		                HXPROC_VERBOSE);
 
 	/* pass on through the result from the umount process */
-	return !WEXITSTATUS(child_exit);
+	return proc.p_exited && proc.p_status == 0;
 }
 
 /**
@@ -706,9 +592,10 @@ int mount_op(mount_op_fn_t *mnt, const struct config *config,
     struct vol *vpt, const char *password)
 {
 	int fnval;
-	struct HXbtree *vinfo;
+	struct HXformat_map *vinfo;
 	struct passwd *pe;
 	hxmc_t *options;
+	char real_mpt[PATH_MAX+1];
 
 	/*
 	 * This expansion (the other is in expandconfig()!) expands the mount
@@ -722,6 +609,16 @@ int mount_op(mount_op_fn_t *mnt, const struct config *config,
 	 */
 	if ((vinfo = HXformat_init()) == NULL)
 		return 0;
+
+	if (realpath(vpt->mountpoint, real_mpt) == NULL) {
+		w4rn("Could not get realpath of %s: %s\n",
+		     vpt->mountpoint, strerror(errno));
+	} else {
+		real_mpt[sizeof(real_mpt)-1] = '\0';
+		free(vpt->mountpoint);
+		vpt->mountpoint = xstrdup(real_mpt);
+	}
+
 	format_add(vinfo, "MNTPT",    vpt->mountpoint);
 	format_add(vinfo, "FSTYPE",   vpt->fstype);
 	format_add(vinfo, "VOLUME",   vpt->volume);
@@ -736,10 +633,10 @@ int mount_op(mount_op_fn_t *mnt, const struct config *config,
 		w4rn("getpwnam(\"%s\") failed: %s\n",
 		     Config.user, strerror(errno));
 	} else {
-		HXformat_add(vinfo, "USERUID", static_cast(void *,
+		HXformat_add(vinfo, "USERUID", reinterpret_cast(void *,
 			static_cast(long, pe->pw_uid)),
 			HXTYPE_UINT | HXFORMAT_IMMED);
-		HXformat_add(vinfo, "USERGID", static_cast(void *,
+		HXformat_add(vinfo, "USERGID", reinterpret_cast(void *,
 			static_cast(long, pe->pw_gid)),
 			HXTYPE_UINT | HXFORMAT_IMMED);
 	}
@@ -763,7 +660,8 @@ int mount_op(mount_op_fn_t *mnt, const struct config *config,
  * Returns 1 if the filesystem does not require a block device, 0 if it does
  * require a block device, -1 if we could not find out.
  */
-static int fstype_nodev(const char *name) {
+int fstype_nodev(const char *name)
+{
 	char buf[80];
 	FILE *fp;
 
@@ -775,8 +673,10 @@ static int fstype_nodev(const char *name) {
 	while (fgets(buf, sizeof(buf), fp) != NULL) {
 		char *bp = buf;
 		HX_chomp(buf);
-		while (isalpha(*bp)) ++bp;
-		while (isspace(*bp)) ++bp;
+		while (!HX_isspace(*bp) && *bp != '\0')
+			++bp;
+		while (HX_isspace(*bp))
+			++bp;
 		if (strcasecmp(bp, name) == 0) {
 			fclose(fp);
 			return strncasecmp(buf, "nodev", 5) == 0;
@@ -796,18 +696,22 @@ void umount_final(struct config *config)
 {
 	struct vol *vol;
 
+	if (HXlist_empty(&config->volume_list.list))
+		/* Avoid needlessy waiting on usleep */
+		return;
+
 	if (config->sig_hup)
 		HXlist_for_each_entry_rev(vol, &config->volume_list, list)
-			ofl(vol->mountpoint, SIGHUP);
+			run_ofl(config, vol->mountpoint, SIGHUP);
 	if (config->sig_term) {
 		usleep(config->sig_wait);
 		HXlist_for_each_entry_rev(vol, &config->volume_list, list)
-			ofl(vol->mountpoint, SIGTERM);
+			run_ofl(config, vol->mountpoint, SIGTERM);
 	}
 	if (config->sig_kill) {
 		usleep(config->sig_wait);
 		HXlist_for_each_entry_rev(vol, &config->volume_list, list)
-			ofl(vol->mountpoint, SIGKILL);
+			run_ofl(config, vol->mountpoint, SIGKILL);
 	}
 	HXlist_for_each_entry_rev(vol, &config->volume_list, list) {
 		w4rn("going to unmount\n");
@@ -816,3 +720,31 @@ void umount_final(struct config *config)
 			    vol->volume);
 	}
 }
+
+/**
+ * fstype_icase - return whether the volume name is case-insensitive
+ * @fstype:	filesystem type (cifs, etc.)
+ *
+ * For some filesystems, notably those which do not distinguish between case
+ * sensitivity, the volume ("share") name is usually also case-insensitive.
+ */
+bool fstype_icase(const char *fstype)
+{
+	if (fstype == NULL)
+		return false;
+	return strcasecmp(fstype, "cifs") == 0 ||
+	       strcasecmp(fstype, "smbfs") == 0 ||
+	       strcasecmp(fstype, "ncpfs") == 0;
+}
+
+bool fstype2_icase(enum command_type fstype)
+{
+	switch (fstype) {
+	case CMD_CIFSMOUNT:
+	case CMD_SMBMOUNT:
+	case CMD_NCPMOUNT:
+		return true;
+	default:
+		return false;
+	}
+}
